from .writer import DataProcessing  # noqa
from .writer import MsInstrument  # noqa
from .writer import MsRun  # noqa
from .writer import MzXml  # noqa
from .writer import ParentFile  # noqa
from .writer import Scan  # noqa



"""Python wrapper for RAMP.

Provides access to the high performance mzXML parsing C library RAMP
via Python.

This module is based on code from the Trans Proteomics Pipeline
(https://sf.net/projects/sashimi)
"""

import ctypes
import os
import sysconfig
from pathlib import Path

import numpy as np

off_t = ctypes.c_long
SCAN_TYPE_LENGTH = 32
INSTRUMENT_LENGTH = 2000
CHARGEARRAY_LENGTH = 128
FILTER_LENGTH = 500

# Do some magic to figure out name and location of ramp library and
# load it
thisfolder = Path(__file__).parent.absolute()
suffix = sysconfig.get_config_var("EXT_SUFFIX")
ramp = ctypes.CDLL(thisfolder.joinpath("ramp" + suffix))


def pairwise(iterable):
    """Returns peak list as m/z - intensity pairs.
    s -> (s0,s1), (s2,s3), ..."""
    item = iter(iterable)
    return zip(item, item)


class SCANHEADERSTRUCT(ctypes.Structure):
    """Python implementation of the RAMP ScanHeaderStruct."""

    _fields_ = [
        ("seqNum", ctypes.c_int),
        ("acquisitionNum", ctypes.c_int),
        ("msLevel", ctypes.c_int),
        ("peaksCount", ctypes.c_int),
        ("totIonCurrent", ctypes.c_double),
        ("retentionTime", ctypes.c_double),
        ("basePeakMz", ctypes.c_double),
        ("basePeakIntensity", ctypes.c_double),
        ("collisionEnergy", ctypes.c_double),
        ("compensationVoltage", ctypes.c_double),
        ("ionisationEnergy", ctypes.c_double),
        ("lowMz", ctypes.c_double),
        ("highMz", ctypes.c_double),
        ("precursorScanNum", ctypes.c_int),
        ("precursorMz", ctypes.c_double),
        ("precursorCharge", ctypes.c_int),
        ("precursorIntensity", ctypes.c_double),
        ("scanType", ctypes.c_char * SCAN_TYPE_LENGTH),
        ("activationMethod", ctypes.c_char * SCAN_TYPE_LENGTH),
        ("possibleCharges", ctypes.c_char * SCAN_TYPE_LENGTH),
        ("numPossibleCharges", ctypes.c_int),
        ("possibleChargesArray", ctypes.c_bool * CHARGEARRAY_LENGTH),
        ("mergedScan", ctypes.c_int),
        ("mergedResultScanNum", ctypes.c_int),
        ("mergedResultStartScanNum", ctypes.c_int),
        ("mergedResultEndScanNum", ctypes.c_int),
        ("filterLine", ctypes.c_char * FILTER_LENGTH),
        ("filePosition", ctypes.c_long),
    ]  # Converted to offset in file from actual RAMPFILE


class RUNHEADERSTRUCT(ctypes.Structure):
    """Python implementation of the RAMP RunHeaderStruct."""

    _fields_ = [
        ("scanCount", ctypes.c_int),
        ("lowMz", ctypes.c_double),
        ("highMz", ctypes.c_double),
        ("startMz", ctypes.c_double),
        ("endMz", ctypes.c_double),
        ("dStartTime", ctypes.c_double),
        ("dEndTime", ctypes.c_double),
    ]


class INSTRUMENTSTRUCT(ctypes.Structure):
    """Python implementation of the RAMP InstrumentStruct."""

    _fields_ = [
        ("manufacturer", ctypes.c_char * INSTRUMENT_LENGTH),
        ("model", ctypes.c_char * INSTRUMENT_LENGTH),
        ("ionisation", ctypes.c_char * INSTRUMENT_LENGTH),
        ("analyzer", ctypes.c_char * INSTRUMENT_LENGTH),
        ("detector", ctypes.c_char * INSTRUMENT_LENGTH),
    ]


class Rampy:
    """Wrapper for the RAMP C library."""

    def __init__(self, mzxml_file_name):
        """Initializations:
        - Opens mzXML file
        - Gets the offset of the index
        - Reads the scan index and store it in mzxml_idx"""
        self.mzxml_file_name = mzxml_file_name
        # Open mzXML file
        self.file_descriptor = self.open_file()
        print("File descriptor #: {}".format(self.file_descriptor))
        # Get the offset of the index
        self.index_offset = self.read_index_offset()
        if self.index_offset == -1:
            print("Problems reading index.")
            print(
                "Please re-generate one with index_mzXML {}".format(
                    mzxml_file_name
                )
            )
            exit()
        else:
            print("mzXML index offset found at: {}".format(self.index_offset))
        self.last_scan_num = (
            ctypes.c_int()
        )  # This is set by reference in read_index()
        # Read the scan index and store it in mzxml_idx
        self.mzxml_idx = self.read_index()
        print("done reading mzXML index")

    def open_file(self):
        """Opens a file returns associated file descriptor.

        RETURNS: a file descriptor

        NOTE: For compatibility with python 3 we open the file in
        python and pass a file descriptor to a wrapper for the RAMP C
        API (which will request a FILE* to it). The wrapper API
        functions have the same name as the original RAMP functions
        with Fd appended at the end. In C RAMP relies on FILE pointers
        as arguments, but these can no longer by passed back to python
        after upgrading from 2 to 3.and. For the same reason we do not
        use the C RAMP rampOpenFile API.
        """
        print("opening {}".format(self.mzxml_file_name))
        fd = os.open(self.mzxml_file_name, os.O_RDONLY)
        return fd

    def close_file(self):
        """Closes the mzXML file associated with the instance of this class.

        NOTE: This method must be called once you are done with the instance.
        """
        os.close(self.file_descriptor)

    def read_index_offset(self):
        """Calls RAMP getIndexOffset() method.

        RETURNS: The file offset (64 bits) of the index element.
        """
        ramp.getIndexOffsetFd.argtypes = [ctypes.c_int]
        ramp.getIndexOffsetFd.restype = off_t
        return ramp.getIndexOffsetFd(self.file_descriptor)

    def read_last_scan_num(self):
        """Calls RAMP getLastScan() method.

        RETURNS: The scan number for the last scan in the mzXML file.
        """
        ramp.getLastScanFd.argtypes = [ctypes.c_int]
        ramp.getLastScanFd.restype = ctypes.c_int
        return ramp.getLastScanFd(self.file_descriptor)

    def read_index(self):
        """Calls RAMP readIndex() method.

        RETURNS: An array with the offsets (64 bits) for the scan in the
        mzXML file.  last_scan_num: will contain the scan number for
        the last scan.

        NOTE: as per mzXML format specifications the scan numbering in the
        file is assumed to be sequential
        """
        # TODO: implement freeing of the index
        tot_scans = self.read_last_scan_num()
        ramp.readIndexFd.argtypes = [
            ctypes.c_int,
            off_t,
            ctypes.POINTER(ctypes.c_int),
        ]
        # We add one since the idx array is 0 based, and the scan
        # numbering is 1 based
        ramp.readIndexFd.restype = ctypes.POINTER(off_t * (tot_scans + 1))
        return ramp.readIndexFd(
            self.file_descriptor,
            self.index_offset,
            ctypes.byref(self.last_scan_num),
        )

    # TODO this is missing in TPP ramp
    #
    # def read_filter_line(self, scan_num):
    #     """Calls RAMP readFilterLine() method.

    #     RETURNS: the filter line the scan scan_num

    #     """
    #     ramp.readFilterLine.argtypes = [FILE_ptr, off_t]
    #     ramp.readFilterLine.restype = None
    #     filter_line = ctypes.create_string_buffer('', 1000)
    #     ramp.readFilterLine(self.file_pointer,self.mzxml_idx.contents[scan_num],
    #                         filter_line)
    #     return filter_line.value

    def read_ms_level(self, scan_num):
        """Calls RAMP readMsLevel() method.

        RETURNS: the MS level (1: MS; 2: MS/MS; ...) for the scan
        scan_num
        """
        ramp.readMsLevelFd.argtypes = [ctypes.c_int, off_t]
        ramp.readMsLevelFd.restype = ctypes.c_int
        return ramp.readMsLevelFd(
            self.file_descriptor, self.mzxml_idx.contents[scan_num]
        )

    def read_peaks_count(self, scan_num):
        """Calls RAMP getPeakscount() method.

        RETURNS: the number of peaks in scan scan_num
        """
        ramp.readPeaksCountFd.argtypes = [ctypes.c_int, off_t]
        ramp.readPeaksCountFd.restype = ctypes.c_int
        return ramp.readPeaksCountFd(
            self.file_descriptor, self.mzxml_idx.contents[scan_num]
        )

    def read_start_mz(self, scan_num):
        """Calls RAMP readStartMz() method.

        RETURNS: The value of startMz for scan scan_num.
                 Will return 1E6 if the value was not set in the mzXML file.
                 This is the value set by the user, not the measured one.
        """
        ramp.readStartMzFd.argtypes = [ctypes.c_int, off_t]
        ramp.readStartMzFd.restype = ctypes.c_double
        return ramp.readStartMzFd(
            self.file_descriptor, self.mzxml_idx.contents[scan_num]
        )

    def read_end_mz(self, scan_num):
        """Calls RAMP readEndMz() method.

        RETURNS: The value of endMz for scan scan_num.
                 Will return 0 if the value was not set in the mzXML file.
                 This is the value set by the user, not the measured one.
        """
        ramp.readEndMzFd.argtypes = [ctypes.c_int, off_t]
        ramp.readEndMzFd.restype = ctypes.c_double
        return ramp.readEndMzFd(
            self.file_descriptor, self.mzxml_idx.contents[scan_num]
        )

    # TODO this is missing in TPP ramp
    #
    # def read_low_mz(self, scan_num):
    #     """Calls RAMP readLowMz() method.

    #     RETURNS: The value of lowMz for scan scan_num.
    #              Will return -1 if the value was not set in the mzXML file.
    #              This is the measured value.
    #     """
    #     ramp.readLowMz.argtypes = [FILE_ptr, off_t]
    #     ramp.readLowMz.restype = ctypes.c_double
    #     return ramp.readLowMz(self.file_pointer,
    #                           self.mzxml_idx.contents[scan_num])

    # TODO this is missing in TPP ramp
    #
    # def read_high_mz(self, scan_num):
    #     """Calls RAMP readHighMz() method.

    #     RETURNS: The value of highMz for scan scan_num.
    #              Will return -1 if the value was not set in the mzXML file.
    #              This is the measured value.
    #     """
    #     ramp.readHighMz.argtypes = [FILE_ptr, off_t]
    #     ramp.readHighMz.restype = ctypes.c_double
    #     return ramp.readHighMz(self.file_pointer,
    #                            self.mzxml_idx.contents[scan_num])

    # TODO this is missing in TPP ramp
    #
    # def read_rt(self, scan_num):
    #     """Calls RAMP readRT() method.

    #     RETURNS: The value of RT for scan scan_num in seconds
    #     """
    #     ramp.readRT.argtypes = [FILE_ptr, off_t]
    #     ramp.readRT.restype = ctypes.c_char_p
    #     rt_duration =  ramp.readRT(self.file_pointer,
    #                                self.mzxml_idx.contents[scan_num])
    #     return float(rt_duration[2:-1])

    def read_run_header(self):
        """Calls RAMP readRunHeader() method.

        RETURNS: a dictionary with the run header information
        """
        ramp.readRunHeaderFd.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(off_t * (self.last_scan_num.value + 1)),
            ctypes.POINTER(RUNHEADERSTRUCT),
            ctypes.c_int,
        ]
        ramp.readRunHeaderFd.restype = None
        run_header_struct = RUNHEADERSTRUCT()
        ramp.readRunHeaderFd(
            self.file_descriptor,
            self.mzxml_idx,
            ctypes.byref(run_header_struct),
            self.last_scan_num,
        )
        return self.struct2dict(run_header_struct)

    def read_ms_run(self):
        """Calls RAMP readMSRun() method.

        RETURNS: a dictionary with the MS run information
        """
        ramp.readMSRunFd.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(RUNHEADERSTRUCT),
        ]
        ramp.readMSRunFd.restype = None
        run_header_struct = RUNHEADERSTRUCT()
        ramp.readMSRunFd(self.file_descriptor, ctypes.byref(run_header_struct))
        return self.struct2dict(run_header_struct)

    # TODO this is most likely leaking memory since RAMP allocates the structure
    def read_instrument_info(self):
        """Calls RAMP getInstrumentStruct() method.

        RETURNS: a dictionary with the instrument information
        """
        ramp.getInstrumentStructFd.argtypes = [ctypes.c_int]
        ramp.getInstrumentStructFd.restype = ctypes.POINTER(INSTRUMENTSTRUCT)
        # instrument_structure = INSTRUMENTSTRUCT.from_address(
        #     ramp.getInstrumentStructFd(self.file_descriptor)
        # )
        instrument_structure = ramp.getInstrumentStructFd(self.file_descriptor)
        if instrument_structure:
            instrument_dict = self.struct2dict(instrument_structure.contents)
            self.free_instrument_struct(instrument_structure)
            return instrument_dict
        return None

    def read_scan_header(self, scan_num):
        """Calls the RAMP readScanHeader() method.

        RETURNS: a dictionary with the scan header information
        """
        ramp.readHeaderFd.argtypes = [
            ctypes.c_int,
            off_t,
            ctypes.POINTER(SCANHEADERSTRUCT),
        ]
        ramp.readHeaderFd.restype = None
        scan_header_struct = SCANHEADERSTRUCT()
        ramp.readHeaderFd(
            self.file_descriptor,
            self.mzxml_idx.contents[scan_num],
            ctypes.byref(scan_header_struct),
        )
        return self.struct2dict(scan_header_struct)

    def read_peaks(self, scan_num, deep_copy=True):
        """Calls RAMP readPeaks() method.

        RETURNS: Returns a list array of (mz - intensity) tuples

        If deep_copy is False returns n array of C floats with the m/z
        - intensity pairs and the frees the peaks memory.
        """
        # TODO: Look into adjusting this function in RAMP to support double if
        # required

        # TODO: Add support for zlib compression type in RAMP
        peaks_count = self.read_peaks_count(scan_num)
        ramp.readPeaksFd.argtypes = [ctypes.c_int, off_t]
        ramp.readPeaksFd.restype = ctypes.POINTER(
            ctypes.c_double * (peaks_count * 2)
        )
        peaks_array = ramp.readPeaksFd(
            self.file_descriptor, self.mzxml_idx.contents[scan_num]
        )
        if deep_copy:
            peaks = pairwise(np.array(peaks_array.contents))
            # Conversion to numpy will also perform a deepcopy for us. We can
            # release the memory
            self.free_peaks_memory(peaks_array)
            return peaks
        else:
            return peaks_array

    @staticmethod
    def free_peaks_memory(pointer_peaks):
        """Frees the memory allocated for the peaks during a read_peaks() call.

        NOTE: This must be called once you are done with pointer_peaks
        to avoid leaks!
        """
        ramp.freePeaks.restype = None
        ramp.freePeaks(pointer_peaks)

    @staticmethod
    def free_instrument_struct(pointer_instrument_struct):
        """Frees the memory allocated for the Instrument Structure during a
        read_instrument_info() call."""

        ramp.freeInstrumentStruct.argtypes = [ctypes.POINTER(INSTRUMENTSTRUCT)]
        ramp.freeInstrumentStruct.restype = None
        ramp.freeInstrumentStruct(pointer_instrument_struct)

    def struct2dict(self, struct):
        return dict((f, getattr(struct, f)) for f, _ in struct._fields_)



from setuptools import Extension, setup

module1 = Extension(
    # The "name" defines where the compiled shared object will be placed
    name="pymzxml.rampy.ramp",
    # Support for large files
    define_macros=[("FILE_OFFSET_BITS", "64")],
    # We need to link against the zlib library
    extra_link_args=["-lz"],
    sources=[
        "pymzxml/ramp/src/ramp.cpp",
        "pymzxml/ramp/src/ramp_base64.cpp",
        "pymzxml/ramp/src/fd_ramp.cpp",
    ],
)

setup(ext_modules=[module1])



"""Collection of classes for writing mzXML files."""

import base64
import hashlib
import re
import struct
import zlib

import pkg_resources


class MzXml:
    """MzXML class to store a full mzXML file."""

    def __init__(self, filename):
        """Initialize the MzXML class."""
        self.filename = filename
        self.file = open(filename, "w")
        self.ms_runs = []
        self.index = []
        self.index_offset = None
        self.sha1 = None

    def __str__(self):
        """Return the mzXML file as a string."""
        mzxml = self.pre_sha1_string()
        self.calculate_sha1_from_string(mzxml)
        mzxml += self.sha1
        mzxml += "</sha1>\n"
        mzxml += self.close_xml()
        return mzxml

    def save_to_file(self):
        """Save the mzXML file to disk."""
        self.file.write(str(self))
        self.file.close()

    def pre_sha1_string(self):
        """Return the string used for the SHA1 checksum.

        This string is the mzXML file up to and including the <sha1> tag
        without the SHA1 checksum.
        """
        mzxml = self.open_xml()
        for ms_run in self.ms_runs:
            mzxml += str(ms_run)
        self.calculate_index_from_string(mzxml)
        mzxml += self.index_to_xml()
        self.calculate_index_offset_from_string(mzxml)
        mzxml += self.index_offset_to_xml()
        mzxml += " <sha1>"
        return mzxml

    @staticmethod
    def open_xml():
        """Return opening XML for an mzXML file.

        Use this instead of __str__ to write to a file in serially.
        """
        return (
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            '<mzXML xmlns="http://sashimi.sourceforge.net/schema_revision/mzXML_3.2"\n'  # noqa
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n'
            ' xsi:schemaLocation="http://sashimi.sourceforge.net/schema_revision/mzXML_3.2'  # noqa
            ' http://sashimi.sourceforge.net/schema_revision/mzXML_3.2/mzXML_idx_3.2.xsd">\n'  # noqa
        )

    @staticmethod
    def close_xml():
        """Return closing XML for an mzXML file.

        Use this instead of __str__ to write file serially.
        """
        return "</mzXML>"

    def add_ms_run(self, ms_run):
        """Add an MsRun to the mzXML file.

        Each MzXml can have multiple MsRuns.
        """
        self.ms_runs.append(ms_run)

    def calculate_index_from_string(self, mzxml_string):
        """Index the file offsets of the scans in the MzXML file."""
        pattern = re.compile(r'<scan num="(\d+)"')
        self.index = [
            (m.start(), m.group(1)) for m in re.finditer(pattern, mzxml_string)
        ]

    def calculate_index_from_file(self):
        """Index the file offsets of the scans in the MzXML file.

        Use this instead of calculate_index_from_string when writing to
        a file in a serial manner.
        """
        with open(self.filename, "r") as mzxml_file:
            pattern = re.compile(r'<scan num="(\d+)"')
            self.index = [
                (m.start(), m.group(1))
                for m in re.finditer(pattern, mzxml_file.read())
            ]

    def calculate_index_offset_from_string(self, mzxml_string):
        """Calculates the offset of the index in the mzXML file."""
        self.index_offset = mzxml_string.find('<index name="scan">')

    def calculate_index_offset_from_file(self):
        """Calculate the offset of the index in the mzXML file.

        Use this instead of calculate_index_offset_from_string when
        writing to a file in a serial manner.
        """
        with open(self.filename, "r") as mzxml_file:
            self.index_offset = mzxml_file.read().find('<index name="scan">')

    def calculate_sha1_from_string(self, pre_sha1_string):
        """Calculate the SHA1 checksum for the mzXML file."""
        sha1 = hashlib.sha1(pre_sha1_string.encode("utf-8"))
        self.sha1 = sha1.hexdigest()

    def calculate_sha1_from_file(self):
        """Calculate the SHA1 checksum for the mzXML file.

        Use this instead of calculate_index_from_string when writing to
        a file in a serial manner.
        """
        with open(self.filename, "r") as mzxml_file:
            sha1 = hashlib.sha1(mzxml_file.read().encode("utf-8"))
        self.sha1 = sha1.hexdigest()

    def index_to_xml(self):
        """Return the index string for the mzXML file."""
        index_string = ' <index name="scan">\n'
        for offset, scan_num in self.index:
            index_string += '  <offset id="{0}">{1}</offset>\n'.format(
                scan_num, offset
            )
        index_string += " </index>\n"
        return index_string

    def index_offset_to_xml(self):
        """Return the index offset string for the mzXML file."""
        return " <indexOffset>{}</indexOffset>\n".format(self.index_offset)


class MsRun:
    """MsRun class to store the information of a single MS run."""

    def __init__(self, scan_count, start_time, end_time):
        """Initialize the MsRun class."""
        self.scan_count = scan_count
        self.start_time = start_time
        self.end_time = end_time
        self.parent_files = []
        self.ms_instrument = None
        self.data_processings = []
        self.scans = []

    def __str__(self):
        """Return the MsRun as a string."""
        ms_run = self.open_xml()
        for parent_file in self.parent_files:
            ms_run += str(parent_file)
        if self.ms_instrument is not None:
            ms_run += str(self.ms_instrument)
        for data_processing in self.data_processings:
            ms_run += str(data_processing)
        for scan in self.scans:
            ms_run += str(scan)
        ms_run += self.close_xml()
        return ms_run

    def open_xml(self):
        """Return opening XML for an MsRun.

        Use this instead of __str__ to write file serially.
        """
        return (' <msRun scanCount="{}" startTime="{}" endTime="{}">\n').format(
            self.scan_count, self.start_time, self.end_time
        )

    @staticmethod
    def close_xml():
        """Return closing XML for an MsRun.

        Use this instead of __str__ to write to a file in serially.
        """
        return " </msRun>\n"

    def update_scan_count(self):
        """Update the scan count of the MsRun.

        This should be called after adding scans to the MsRun.  Note
        that it is not called automatically when writing an mzXML file
        via __str__ method to allow for adding scans in series.
        """
        self.scan_count = len(self.scans)

    def add_parent_file(self, parent_file):
        """Add a parent file to the MsRun.

        Each MsRun can have multiple parent files.
        """
        self.parent_files.append(ParentFile(parent_file))

    def add_ms_instrument(self, ms_instrument):
        """Add an MsInstrument to the MsRun."""
        self.ms_instrument = ms_instrument

    def add_data_processing(
        self, software, software_version, software_type, **kwargs
    ):
        """Add a DataProcessing to the MsRun.

        Each MsRun can have multiple DataProcessings steps.
        """
        self.data_processings.append(
            DataProcessing(
                software=software,
                software_version=software_version,
                software_type=software_type,
                **kwargs
            )
        )

    def add_scan(self, scan):
        """Add a Scan to the MsRun.

        Each MsRun can have multiple Scans.
        """
        self.scans.append(scan)

    def renumber_scans(self, start=1):
        """Renumber the scans in the MsRun."""
        i = start
        for scan in self.scans:
            scan.scan_num = i
            i += 1
            i = scan.renumber_scans(i)


class ParentFile:
    """ParentFile class to store the information of a parent file."""

    def __init__(self, filename):
        """Initialize the ParentFile class."""
        self.filename = filename
        self.filetype = self.get_filetype()
        self.sha1 = self.get_sha1()

    def __str__(self):
        """Return the ParentFile as a  string."""
        return (
            '  <parentFile fileName="{}" fileType="{}" fileSha1="{}"/>\n'
        ).format(self.filename, self.filetype, self.sha1)

    def get_filetype(self):
        """Check the filetype of the parent file.

        Currently only supports mzXML and RAW.
        """
        if self.filename.endswith(".mzXML"):
            return "processedData"
        if self.filename.endswith(".RAW"):
            return "RAWData"
        raise ValueError("Filetype not supported.")

    def get_sha1(self):
        """Calculate the SHA1 hash of the parent file."""
        with open(self.filename, "rb") as file:
            data = file.read()
        return hashlib.sha1(data).hexdigest()


class MsInstrument:
    """MsInstrument class to store the information of the MS instrument."""

    def __init__(
        self,
        ms_manufacturer=None,
        ms_model=None,
        ms_ionisation=None,
        ms_mass_analyzer=None,
        ms_detector=None,
    ):
        """Initialize the MsInstrument class."""
        self.ms_manufacturer = ms_manufacturer
        self.ms_model = ms_model
        self.ms_ionisation = ms_ionisation
        self.ms_mass_analyzer = ms_mass_analyzer
        self.ms_detector = ms_detector

    def __str__(self):
        """Return the MsInstrument as a string."""
        ms_instrument = "  <msInstrument>\n"
        if self.ms_manufacturer is not None:
            ms_instrument += (
                '   <msManufacturer category="msManufacturer" value="{}"/>\n'
            ).format(self.ms_manufacturer)
        if self.ms_model is not None:
            ms_instrument += (
                '   <msModel category="msModel" value="{}"/>\n'
            ).format(self.ms_model)
        if self.ms_ionisation is not None:
            ms_instrument += (
                '   <msIonisation category="msIonisation" value="{}"/>\n'
            ).format(self.ms_ionisation)
        if self.ms_mass_analyzer is not None:
            ms_instrument += (
                '   <msMassAnalyzer category="msMassAnalyzer" value="{}"/>\n'
            ).format(self.ms_mass_analyzer)
        if self.ms_detector is not None:
            ms_instrument += (
                '   <msDetector category="msDetector" value="{}"/>\n'
            ).format(self.ms_detector)
        ms_instrument += "  </msInstrument>\n"
        return ms_instrument


class DataProcessing:
    """DataProcessing class to store the information of the data processing."""

    def __init__(
        self,
        software="pymzxml",
        software_version=pkg_resources.get_distribution("pymzxml").version,
        software_type="processing",
        intensityCutoff=None,
        centroided=None,
        deisotoped=None,
        chargeDeconvoluted=None,
        spotIntegration=None,
    ):
        """Initialize the DataProcessing class."""
        self.software = software
        self.software_version = software_version
        self.software_type = software_type
        self.intensityCutoff = intensityCutoff
        self.centroided = centroided
        self.deisotoped = deisotoped
        self.chargeDeconvoluted = chargeDeconvoluted
        self.spotIntegration = spotIntegration

    def __str__(self):
        """Return the DataProcessing as a string."""
        data_processing = "  <dataProcessing"
        if self.intensityCutoff is not None:
            data_processing += 'intensityCutoff="{}"'.format(
                self.intensityCutoff
            )
        if self.centroided is not None:
            if type(self.centroided) is not bool:
                raise ValueError("centroided must be boolean.")
            data_processing += ' centroided="{}"'.format(int(self.centroided))
        if self.deisotoped is not None:
            if type(self.deisotoped) is not bool:
                raise ValueError("deisotoped must be boolean.")
            data_processing += ' deisotoped="{}"'.format(int(self.deisotoped))
        if self.chargeDeconvoluted is not None:
            data_processing += ' chargeDeconvoluted="{}"'.format(
                self.chargeDeconvoluted
            )
        if self.spotIntegration is not None:
            data_processing += ' spotIntegration="{}"'.format(
                self.spotIntegration
            )
        data_processing += ">\n"
        data_processing += (
            '   <software type="{}" name="{}" version="{}"/>\n'
        ).format(self.software_type, self.software, self.software_version)
        data_processing += "  </dataProcessing>\n"
        return data_processing


class Scan:
    """Scan class to store information about a single scan."""

    def __init__(
        self,
        scan_num,
        ms_level,
        retention_time,
        peaks,
        polarity=None,
        scan_type=None,
        filter_line=None,
        injection_time=None,
        precursor=None,
        precursor_intensity=None,
        activation_method=None,
        window_wideness=None,
        precursor_charge=None,
        precision=32,
        compress=False,
    ):
        """Initialize the Scan class."""
        self.scan_num = scan_num
        self.peaks_count = int(len(peaks) / 2)
        self.polarity = polarity
        self.scan_type = scan_type
        self.filter_line = filter_line
        self.injection_time = injection_time
        self.low_mz = min(peaks[::2])
        self.high_mz = max(peaks[::2])
        self.base_peak_mz = peaks[peaks.index(max(peaks[1::2])) - 1]
        self.base_peak_intensity = max(peaks[1::2])
        self.tot_ion_current = sum(peaks[1::2])
        self.ms_level = ms_level
        self.retention_time = retention_time
        self.precursor = precursor
        self.precursor_intensity = precursor_intensity
        self.activation_method = activation_method
        self.window_wideness = window_wideness
        self.precursor_charge = precursor_charge
        self.precision = precision
        self.compression_type = "zlib" if compress else "none"
        self.compressed_len = None
        self.peaks = self.encode_peaks(peaks, precision, compress)
        self.scans = []

    def __str__(self):
        """Return the Scan as a string."""
        scan = self.open_xml()
        for sub_scan in self.scans:
            scan += str(sub_scan)
        scan += self.close_xml()
        return scan

    def open_xml(self):
        """Return the opening XML tags of the Scan.

        Use this instead of __str__ to write to a file serially.
        """
        scan_element = (
            '  <scan num="{}"\n'
            '   msLevel="{}"\n'
            '   peaksCount="{}"\n'
            '   retentionTime="PT{}S"\n'
            '   lowMz="{}"\n'
            '   highMz="{}"\n'
            '   basePeakMz="{}"\n'
            '   basePeakIntensity="{}"\n'
            '   totIonCurrent="{}"'
        ).format(
            self.scan_num,
            self.ms_level,
            self.peaks_count,
            self.retention_time,
            self.low_mz,
            self.high_mz,
            self.base_peak_mz,
            self.base_peak_intensity,
            self.tot_ion_current,
        )
        if self.polarity is not None:
            scan_element += '\n   polarity="{}"'.format(self.polarity)
        if self.scan_type is not None:
            scan_element += '\n   scanType="{}"'.format(self.scan_type)
        if self.filter_line is not None:
            scan_element += '\n   filterLine="{}"'.format(self.filter_line)
        if self.injection_time is not None:
            scan_element += '\n   injectionTime="{}"'.format(
                self.injection_time
            )
        scan_element += ">\n"
        if self.precursor is not None:
            scan_element += "   <precursorMz"
            if self.precursor_intensity is not None:
                scan_element += ' precursorIntensity="{}"'.format(
                    self.precursor_intensity
                )
            if self.activation_method is not None:
                scan_element += ' activationMethod="{}"'.format(
                    self.activation_method
                )
            if self.window_wideness is not None:
                scan_element += ' windowWideness="{}"'.format(
                    self.window_wideness
                )
            if self.precursor_charge is not None:
                scan_element += ' precursorCharge="{}"'.format(
                    self.precursor_charge
                )
            scan_element += ">{}</precursorMz>\n".format(self.precursor)
        scan_element += (
            '   <peaks precision="{}"\n'
            '    byteOrder="network"\n'
            '    compressionType="{}"'
        ).format(self.precision, self.compression_type)
        if self.compressed_len is not None:
            scan_element += '\n    compressedLen="{}"'.format(
                self.compressed_len
            )
        scan_element += ">{}</peaks>\n".format(self.peaks.decode("utf-8"))
        return scan_element

    @staticmethod
    def close_xml():
        """Return the closing XML tags of the Scan.

        Use this instead of __str__ to write to a file serially.
        """
        return "  </scan>\n"

    def encode_peaks(self, peaks, precision=32, compress=False):
        """Encode peaks to network byte order and base64 encode them."""
        binary_writer = struct.Struct(
            ">{}".format("f" if precision == 32 else "d")
        )
        binary_peaks = bytes()
        for value in peaks:
            binary_peaks += binary_writer.pack(value)
        if compress:
            binary_peaks = zlib.compress(binary_peaks)
            self.compressed_len = len(binary_peaks)
        return base64.b64encode(binary_peaks)

    @staticmethod
    def decode_peaks(byte_peaks, precision=32, compress=False):
        """Decode base64 encoded peaks fix endianess and convert them to
        float."""
        binary_reader = struct.Struct(
            ">{}".format("f" if precision == 32 else "d")
        )
        peaks = []
        binary_peaks = base64.b64decode(byte_peaks)
        if compress:
            binary_peaks = zlib.decompress(binary_peaks)
        for _, k in enumerate(binary_reader.iter_unpack(binary_peaks)):
            peaks += [k[0]]
        return peaks

    def add_scan(self, scan):
        """Add a scan to the list of scans."""
        self.scans.append(scan)

    def renumber_scans(self, start=1):
        """Renumber the scans in the list of scans."""
        i = start
        for scan in self.scans:
            scan.scan_num = i
            i += 1
            i = scan.renumber_scans(i)
        return i
